import pytest
import json
from unittest.mock import MagicMock, patch
from datetime import datetime
from container_hub.carriers.marathon.backend import MarathonBackend
from container_hub.models import (
    LogLevel,
    MountPoint,
    MarathonBackendConfig,
    ContainerConfig,
    EnvVar,
    Label,
    MarathonConstraint,
)


@pytest.fixture
def marathon_backend():
    config = MarathonBackendConfig(
        "http://marathon_url/",
        "my_network",
        constraints=[MarathonConstraint("api", "CLUSTER", "v3")],
    )
    return MarathonBackend(config)


def test_container_hosts(marathon_backend: MarathonBackend):
    with patch("container_hub.carriers.marathon.backend.MarathonClient") as client:
        task = MagicMock()
        task.app_id = "simulation-112"
        task.host = "localhost"
        client().list_tasks.return_value = [task]
        hosts = marathon_backend.container_hosts()
        assert hosts == {"simulation-112": "localhost"}


def test_container_list(marathon_backend: MarathonBackend):
    with patch("container_hub.carriers.marathon.backend.MarathonClient") as client:
        container = MagicMock()
        container.labels = {"simulation_id": "112"}
        client().list_apps.return_value = [container]
        containers = marathon_backend.container_list()
        assert containers == [
            "112",
        ]


def test_container_ips(marathon_backend: MarathonBackend):
    with patch("container_hub.carriers.marathon.backend.MarathonClient") as client:
        task = MagicMock()
        task.app_id = "simulation-112"
        ip_address = MagicMock()
        ip_address.ip_address = "127.0.0.1"
        task.ip_addresses = [ip_address]
        client().list_tasks.return_value = [task]
        ip_addresses = marathon_backend.container_ips()
        assert ip_addresses == {"simulation-112": "127.0.0.1"}


def test_up(marathon_backend: MarathonBackend):
    dt = datetime.now()
    container_config = ContainerConfig(
        "my_image",
        "base_result_path",
        12,
        dt,
        3600,
        3600,
        0,
        "initialize",
        "/model.ini",
        2,
        512,
        [EnvVar("env", "1")],
        [Label("name", "value")],
        0,
        True,
        "gridadmin_url",
        "tables_download_url",
        [MountPoint("/local", "/mnt", False)],
        "redis1",
        LogLevel.debug,
    )

    with patch("container_hub.carriers.marathon.backend.MarathonClient") as client:
        app = MagicMock()
        app.id = "simulation-112"
        client().create_app.return_value = app
        app_id = marathon_backend.up(container_config)
        assert app_id == "simulation-112"

        # Check call args
        to_check = {
            "args": [
                "python",
                "service.py",
                "redis1",
                "/model.ini",
                12,
                dt.isoformat(),
                "3600",
                "3600",
                "initialize",
                "0",
                "0",
                "True",
                "gridadmin_url",
                "tables_download_url",
            ],
            "constraints": [["api", "CLUSTER", "v3"]],
            "container": {
                "docker": {
                    "forcePullImage": False,
                    "image": "my_image",
                    "network": "BRIDGE",
                    "parameters": [{"key": "network", "value": "my_network"}],
                    "privileged": False,
                },
                "type": "DOCKER",
                "volumes": [
                    {"containerPath": "/mnt", "hostPath": "/local", "mode": "RW"}
                ],
            },
            "cpus": 2,
            "env": {
                "LOG_LEVEL": "DEBUG",
                "RESULT_PATH": "base_result_path/simulation-12",
                "env": "1",
            },
            "labels": {"name": "value", "simulation_id": "12"},
            "mem": 512,
        }

        assert client().create_app.call_args[0][1].to_json() == json.dumps(to_check)


def test_down(marathon_backend: MarathonBackend):
    with patch("container_hub.carriers.marathon.backend.MarathonClient") as client:
        app = MagicMock()
        app.id = "simulation-112"
        client().list_apps.return_value = [app]
        marathon_backend.down("112")



import logging
from container_hub.models import DockerBackendConfig, ContainerConfig
from docker import DockerClient
from docker.types import Mount
from docker.errors import DockerException
from docker.errors import APIError
from docker.errors import ContainerError
from docker.errors import NotFound
from functools import cached_property
from typing import List, Dict
from pathlib import Path

from container_hub.exceptions import CarrierError

logger = logging.getLogger(__name__)


class DockerBackend:
    """
    Backend for starting Docker instances via Docker
    """

    def __init__(self, config: DockerBackendConfig):
        self.config = config

    @cached_property
    def client(self):
        return DockerClient(base_url=self.config.client_url)

    def container_hosts(self) -> Dict[str, str]:
        """
        Always localhost, so return emtpy dict
        """
        return {}

    def container_list(self) -> List[str]:
        """
        Returns a list of simulation_ids
        """
        lc = self.client.containers.list(filters={"label": "simulation_id"})
        return [x.name.lstrip("simulation-") for x in lc]

    def container_ips(self) -> Dict[str, str]:
        """
        Return list of ip addresses
        """
        d = {}
        containers = self.client.containers.list(filters={"label": "simulation_id"})
        for container in containers:
            try:
                ip_address = container.attrs["NetworkSettings"]["Networks"][
                    self.config.network_name
                ]["IPAddress"]
            except KeyError:
                continue
            d[container.name] = ip_address
        return d

    def up(self, container_config: ContainerConfig) -> str:
        """
        Create container based on simulation and threedimodel.

        :returns the container id
        """
        name = f"simulation-{container_config.sim_uid}"
        result_path = container_config.base_result_path / Path(name)
        labels = dict([(x.name, x.value) for x in container_config.labels])
        labels.update({"simulation_id": f"{container_config.sim_uid}"})

        _envs = dict([(f"{x.name}", f"{x.value}") for x in container_config.envs])
        _envs.update({"RESULT_PATH": f"{result_path.as_posix()}"})

        if (
            container_config.container_log_level is not None
            and "LOG_LEVEL" not in _envs
        ):
            _envs.update(
                {f"LOG_LEVEL": f"{container_config.container_log_level.value}"}
            )

        if container_config.debugpy_port is not None:
            ports = {
                f"{container_config.debugpy_port}/tcp": container_config.debugpy_port
            }
            _envs.update(
                {f"DEBUGPY": "1", "DEBUGPY_PORT": str(container_config.debugpy_port)}
            )
        else:
            ports = {}

        cmd = (
            f"python service.py {container_config.redis_host} {container_config.model_config} "
            f"{container_config.sim_uid} {container_config.sim_ref_datetime.isoformat()} "
            f"{container_config.end_time} {container_config.duration} {container_config.start_mode} "
            f"{container_config.pause_timeout} {container_config.max_rate} {container_config.clean_up_files}"
        )

        skip_model_mount: bool = all(
            [
                x
                for x in [
                    container_config.gridadmin_download_url,
                    container_config.tables_download_url,
                ]
            ]
        )

        if container_config.gridadmin_download_url is not None:
            cmd += f" {container_config.gridadmin_download_url}"
        if container_config.tables_download_url is not None:
            cmd += f" {container_config.tables_download_url}"

        logger.debug("cmd %s", cmd)
        logger.debug("Envs %s", _envs)

        mounts: List[Mount] = []
        for mount in container_config.mount_points:
            if skip_model_mount and mount.mount_path == "/models":
                # Skip mounting models
                continue
            mounts.append(
                Mount(
                    mount.mount_path,
                    mount.local_path,
                    type="bind",
                    read_only=mount.read_only,
                )
            )

        try:
            container = self.client.containers.run(
                image=container_config.image_name,
                command=cmd,
                name=name,
                network=self.config.network_name,
                mounts=mounts,
                environment=_envs,
                detach=True,
                labels=labels,
                ports=ports,
            )
        except (DockerException, APIError, ContainerError) as err:
            logger.error(err)
            raise CarrierError(err)

        # double check everything went right
        try:
            self.client.containers.get(container.id)
        except (APIError, NotFound) as err:
            logger.error(
                f"simulation container exited prematurely. Could not retrieve "
                f"the container if though it should be running {err}"
            )
            raise CarrierError(err)

        logger.info(f"Started simulation container {container.name}")
        return container.id

    def down(self, sim_uid: str):
        try:
            container = self.client.containers.get(f"simulation-{sim_uid}")
            container_id = container.id
        except (APIError, NotFound) as err:
            logger.error(
                f"Could not get the simulation container, error message: {err}"
            )
            raise CarrierError(err)
        try:
            if not self.config.debug:
                container.remove(force=True)
            else:
                container.kill()
        except APIError as err:
            logger.error(
                f"Could not kill/remove the "
                f"simulation container, error message: {err}"
            )
            raise CarrierError(err)
        logger.info(f"Removed container for simulation {sim_uid}")
        return container_id



import logging
from pathlib import Path
from typing import Dict, List
from typing import Iterator
from typing import Tuple
from container_hub.models import MarathonBackendConfig
from container_hub.models import ContainerConfig
from functools import cached_property
import marathon.exceptions
from marathon import MarathonApp
from marathon.models.app import MarathonTask
from marathon.models.container import MarathonContainer
from marathon.models.container import MarathonContainerVolume
from marathon.models.container import MarathonDockerContainer
from marathon.models.constraint import MarathonConstraint
from marathon import MarathonClient
from container_hub.exceptions import CarrierError

logger = logging.getLogger(__name__)


class MachineManagerException(Exception):
    pass


def _task_generator(client: MarathonClient) -> Iterator[Tuple[MarathonTask, str]]:
    tasks = client.list_tasks()
    for task in tasks:
        app_name = task.app_id.strip("/")
        yield task, app_name


def _del_app(client: MarathonClient, app_name: str) -> bool:
    """
    wrapper around the ``client.delete_app`` call

    :returns True if the app has been deleted, False otherwise
    """
    try:
        client.delete_app(app_name, force=True)
    except marathon.exceptions.NotFoundError:
        # not found; assume that the app is deleted
        logger.exception("Failed to delete app %s.", app_name)
        return False

    logger.info("Deleted app %s.", app_name)
    return True


class MarathonBackend:
    """
    Backend for starting Docker instances via Docker
    """

    def __init__(self, config: MarathonBackendConfig):
        self.config = config

    @cached_property
    def client(self) -> MarathonClient:
        return MarathonClient(self.config.client_url)

    def container_hosts(self) -> Dict[str, str]:
        d = {}
        for task, app_name in _task_generator(self.client):
            d[app_name] = task.host
        return d

    def container_ips(self) -> Dict[str, str]:
        d = {}
        for task, app_name in _task_generator(self.client):
            _ip = task.ip_addresses[0]
            d[app_name] = _ip.ip_address
        return d

    def container_list(self) -> List[str]:
        sim_uids = []
        apps = self.client.list_apps()
        for app in apps:
            # v3 apps all have a simulation id label
            sim_uid = app.labels.get("simulation_id")
            if not sim_uid:
                continue
            sim_uids.append(sim_uid)
        return sim_uids

    def up(self, container_config: ContainerConfig) -> str:
        """Create a MarathonApp instance."""
        name = f"simulation-{container_config.sim_uid}"

        labels = dict([(x.name, x.value) for x in container_config.labels])
        labels.update({"simulation_id": f"{container_config.sim_uid}"})

        docker_container = MarathonDockerContainer(
            image=container_config.image_name,
            network="BRIDGE",
            parameters=[{"key": "network", "value": self.config.network_name}],
        )

        volumes = []  # == mounts

        skip_model_mount: bool = all(
            [
                x
                for x in [
                    container_config.gridadmin_download_url,
                    container_config.tables_download_url,
                ]
            ]
        )

        for mount in container_config.mount_points:
            if skip_model_mount and mount.mount_path == "/models":
                # Skip mounting models
                continue

            volumes.append(
                MarathonContainerVolume(
                    container_path=mount.mount_path,
                    host_path=mount.local_path,
                    mode="RO" if mount.read_only else "RW",
                )
            )

        logger.debug("Volumes %s", volumes)

        # docker container with volumes
        container = MarathonContainer(docker=docker_container, volumes=volumes)
        result_path = container_config.base_result_path / Path(name)

        envs = dict([(f"{x.name}", f"{x.value}") for x in container_config.envs])
        # environment variables for container
        envs.update({"RESULT_PATH": result_path.as_posix()})

        if container_config.container_log_level is not None and "LOG_LEVEL" not in envs:
            envs.update({"LOG_LEVEL": f"{container_config.container_log_level.value}"})

        constraints = []

        if self.config.constraints:
            constraints = [
                MarathonConstraint(
                    constraint.param, constraint.operator, constraint.value
                )
                for constraint in self.config.constraints
            ]

        # all args must be strings
        args = [
            "python",
            "service.py",
            container_config.redis_host,
            container_config.model_config,
            container_config.sim_uid,
            container_config.sim_ref_datetime.isoformat(),
            str(container_config.end_time),
            str(container_config.duration),
            container_config.start_mode,
            str(container_config.pause_timeout),
            str(container_config.max_rate),
            str(container_config.clean_up_files),
        ]

        if container_config.gridadmin_download_url is not None:
            args.append(str(container_config.gridadmin_download_url))
        if container_config.tables_download_url is not None:
            args.append(str(container_config.tables_download_url))

        marathon_app_definition = MarathonApp(
            args=args,
            container=container,
            mem=container_config.session_memory,
            cpus=container_config.max_cpu,
            env=envs,
            labels=labels,
            constraints=constraints,
        )

        try:
            app = self.client.create_app(name, marathon_app_definition)
        except marathon.exceptions.MarathonHttpError as err:
            logger.exception("Failed to create app %s with error %s", name, err)
            raise CarrierError(err)
        logger.info(f"App {app.id} started")
        return app.id

    def down(self, sim_uid: str):
        """Remove the given app."""

        app_name = f"simulation-{sim_uid}"

        # should return a single marathon.models.app.MarathonApp instance
        apps = self.client.list_apps(app_id=app_name)
        if not apps:
            logger.warning(
                "App not found; assuming that the app has already been deleted"
            )
            return

        if len(apps) > 1:
            msg = (
                f"Found more than one app that matches the name {app_name}. "
                f"Trying to delete all of them now..."
            )
            logger.warning(msg)
            for app in apps:
                _del_app(self.client, app.id)
            return

        app = apps[0]
        _del_app(self.client, app.id)



CONTAINER_HUB_CARRIER = "marathon"
CONTAINER_HUB_CLIENT_URL = "http://marathon_url/"
CONTAINER_HUB_NETWORK_NAME = "threedi_backend"
CONTAINER_HUB_IMAGE_NAME = "testimage"
CONTAINER_HUB_CONTAINER_LOG_LEVEL = "DEBUG"
CONTAINER_HUB_MAX_CPU = 2

CONTAINER_HUB_REDIS_HOST = "redis"
CONTAINER_HUB_BASE_MODEL_PATH = "/models"
CONTAINER_HUB_BASE_RESULT_PATH = "/results"

CONTAINER_HUB_MOUNT_POINTS = {
    "local_path_1": {"bind": "mount_path_1", "ro": True},
    "local_path_2": {"bind": "mount_path_2", "ro": False},
}

COMTAINER_HUB_CONSTRAINTS = [
    ["api", "CLUSTER", "v3"],
]
